@@ -224,7 +224,8 @@ GnomeFontFace *GPOFontMap::getFontFaceEmbedded(GfxFont *font)
 
   strObj.streamReset();
   if (!getStreamContents(strObj.getStream(), &contents, &length))
-    g_assert_not_reached(); /* FIXME */
+    return getFontFaceFallback(font);
+
   strObj.free();
 
   switch (font->getType()) {
@@ -262,7 +263,9 @@ GnomeFontFace *GPOFontMap::getFontFaceEmbedded(GfxFont *font)
     FILE *f;
     gushort *code_to_gid;
 
-    ff = FoFiTrueType::make((char *)contents, length); // FIXME error handling
+    ff = FoFiTrueType::make((char *)contents, length);
+    if (!ff)
+      return getFontFaceFallback(font);
 
     code_to_gid = ((Gfx8BitFont *)font)->getCodeToGIDMap(ff); // this is g(oo)malloc'd
 
@@ -303,6 +306,9 @@ GnomeFontFace *GPOFontMap::getFontFaceEmbedded(GfxFont *font)
     gushort *code_to_gid;
     
     ff = FoFiType1C::make((char *)contents, length);
+    if (!ff)
+      return getFontFaceFallback(font);
+
     code_to_gid = ff->getCIDToGIDMap(&n_cids); // this is g(oo)malloc'd
     delete ff;
 
@@ -325,7 +331,9 @@ GnomeFontFace *GPOFontMap::getFontFaceEmbedded(GfxFont *font)
     gint n_cids;    
     gushort *code_to_gid;
 
-    ff = FoFiTrueType::make((char *)contents, length); // FIXME error handling
+    ff = FoFiTrueType::make((char *)contents, length);
+    if (!ff)
+      return getFontFaceFallback(font);
 
     fd = g_file_open_tmp("gpdf-ttf-XXXXXX", &temp_name, NULL);
     f = fdopen(fd, "wb");
