@@ -17,6 +17,7 @@
 #include <time.h>
 #include <unistd.h>
 #include "gpdf-control.h"
+#include "gpdf-control-private.h"
 #include "gpdf-stock-icons.h"
 #include "gpdf-view.h"
 #include "gpdf-bookmarks-view.h"
@@ -88,6 +89,10 @@ struct _GPdfControlPrivate {
 	
 	BonoboControl *bonobo_page_control;
 
+	GtkWidget *file_chooser;
+	GtkWidget *properties;
+	GtkWidget *print_dialog;
+
 	guint idle_id;
 };
 
@@ -422,10 +427,14 @@ gpdf_control_private_get_annots_view (GPdfControl * control)
 }
 #endif
 
-static void
-gpdf_control_private_destroy_dialog (GtkWidget *w, GtkWidget *dlg)
+void
+destroy_widget_unref_control (GtkWidget *widget, gint arg1, BonoboControl *control)
 {
-	gtk_widget_destroy (dlg); 
+	g_return_if_fail (GTK_IS_WIDGET (widget));
+	g_return_if_fail (BONOBO_IS_CONTROL (control));
+
+	gtk_widget_destroy (widget);
+	bonobo_object_unref (BONOBO_OBJECT (control));
 }
 
 /* Create a HIG compliant dialog, display it and
@@ -434,11 +443,15 @@ void
 gpdf_control_private_error_dialog (GPdfControl * control,
 				   const gchar *header_text,
 				   const gchar *body_text,
-				   gboolean modal)
+				   gboolean modal,
+				   gboolean ref_parent)
 {
 	GtkWidget *dlg, *but;
 
 	g_return_if_fail (GPDF_IS_NON_NULL_CONTROL (control));
+	if (ref_parent)
+		bonobo_object_ref (BONOBO_OBJECT (control));
+
 	dlg = gpdf_hig_dialog_new (GTK_STOCK_DIALOG_ERROR,
 				   header_text,
 				   body_text,
@@ -452,11 +465,14 @@ gpdf_control_private_error_dialog (GPdfControl * control,
 	if (modal) {
 		gtk_dialog_run (GTK_DIALOG (dlg));
 		gtk_widget_destroy (dlg);
-	}
-	else {
-		g_signal_connect (G_OBJECT (but), "clicked",
-				  G_CALLBACK (gpdf_control_private_destroy_dialog),
-				  GTK_WIDGET (dlg));
+		if (ref_parent)
+			bonobo_object_unref (BONOBO_OBJECT (control));
+	} else {
+		g_signal_connect (G_OBJECT (dlg), "response",
+				  (ref_parent
+				   ? G_CALLBACK (destroy_widget_unref_control)
+				   : G_CALLBACK (gtk_widget_destroy)),
+				  control);
 		gtk_widget_show (dlg);
 	}		
 }
@@ -483,11 +499,10 @@ gpdf_control_private_warn_dialog (GPdfControl * control,
 	if (modal) {
 		gtk_dialog_run (GTK_DIALOG (dlg));
 		gtk_widget_destroy (dlg);
-	}
-	else {
-		g_signal_connect (G_OBJECT (but), "clicked",
-				  G_CALLBACK (gpdf_control_private_destroy_dialog),
-				  GTK_WIDGET (dlg));
+	} else {
+		g_signal_connect (G_OBJECT (dlg), "response",
+				  G_CALLBACK (gtk_widget_destroy),
+				  NULL);
 		gtk_widget_show (dlg);
 	}
 }
@@ -514,11 +529,9 @@ gpdf_control_private_info_dialog (GPdfControl * control,
 	if (modal) {
 		gtk_dialog_run (GTK_DIALOG (dlg));
 		gtk_widget_destroy (dlg);
-	}
-	else {
-		g_signal_connect (G_OBJECT (but), "clicked",
-				  G_CALLBACK (gpdf_control_private_destroy_dialog),
-				  GTK_WIDGET (dlg));
+	} else {
+		g_signal_connect (G_OBJECT (dlg), "response",
+				  G_CALLBACK (gtk_widget_destroy), NULL);
 		gtk_widget_show (dlg);		
 	}
 }
@@ -545,6 +558,7 @@ gpdf_control_private_display_help (BonoboControl *control,
 		gpdf_control_private_error_dialog (gpdf_control,
 						   _("Documentation Error"),
 						   msg,
+						   FALSE,
 						   FALSE);
 	}
 }
@@ -757,7 +771,7 @@ idle_print_handler (GPdfPrintJob *pj)
 
 		pj->pdf_doc->displayPage (pj->ps_output_dev,
 					  pj->page_next,
-					  72, 0, gFalse);
+					  72.0, 72.0, 0, gTrue, gFalse);
 		pj->page_next++;
 		return TRUE;
 	} else {
@@ -879,17 +893,14 @@ gpdf_control_no_postscript_printer_alert (GPdfControl *control,
 	else
 		secondary = g_strdup (_("You were trying to print to a printer that does not use PostScript. This program requires a PostScript printer driver."));
 
-	dlg = gpdf_hig_dialog_new (
-		GTK_STOCK_DIALOG_ERROR,
+	gpdf_control_private_error_dialog (
+		control,
 		_("Printing is not supported on this printer."),
-		secondary, FALSE);
+		secondary,
+		FALSE,
+		FALSE);
+
 	g_free (secondary);
-	gtk_dialog_add_button (GTK_DIALOG (dlg), GTK_STOCK_OK, GTK_RESPONSE_OK);
-	g_signal_connect_swapped (G_OBJECT (dlg), "response",
-				  G_CALLBACK (gtk_widget_destroy), dlg);
-	bonobo_control_set_transient_for (BONOBO_CONTROL (control),
-					  GTK_WINDOW (dlg), NULL);
-	gtk_widget_show (dlg);
 }
 
 static void
@@ -965,7 +976,7 @@ gpdf_control_print (GPdfControl *control)
 	GPdfPrintJob *pj;
 	GnomePrintConfig *config;
 	GnomePrintJob *job;
-	GtkWidget *print_dialog;
+	GPdfControlPrivate *priv;
 	gint response;
 	GdkWindow *view_window;
 	GdkDisplay *display;
@@ -977,9 +988,11 @@ gpdf_control_print (GPdfControl *control)
 	config = gnome_print_config_default ();
 	job = gnome_print_job_new (config);
 
-	print_dialog = gpdf_control_print_dialog_new (control, job);
-	response = gtk_dialog_run (GTK_DIALOG (print_dialog));
-	gtk_widget_hide (print_dialog);
+	priv = control->priv;
+	
+	priv->print_dialog = gpdf_control_print_dialog_new (control, job);
+	response = gtk_dialog_run (GTK_DIALOG (priv->print_dialog));
+	gtk_widget_hide (priv->print_dialog);
 
 	if (response != GNOME_PRINT_DIALOG_RESPONSE_PRINT)
 		goto exit;
@@ -992,10 +1005,11 @@ gpdf_control_print (GPdfControl *control)
 	pj = g_new0 (GPdfPrintJob, 1);
 
 	gpdf_control_get_range_page (
-		control, GNOME_PRINT_DIALOG (print_dialog),
+		control, GNOME_PRINT_DIALOG (priv->print_dialog),
 		&pj->page_start, &pj->page_end);
 
-	gtk_widget_destroy (print_dialog);
+	gtk_widget_destroy (priv->print_dialog);
+	priv->print_dialog = NULL;
 
 	print_to_file_workaround (job);
 
@@ -1016,86 +1030,184 @@ exit:
 	gpdf_control_enable_ui (control, TRUE);
 }
 
-static void
-gpdf_control_properties_dialog_response_cb (GtkDialog *dialog, gint response_id,
-					    gpointer data)
+static gboolean
+overwrite_existing_file (GtkWindow *window, const gchar *file_name)
 {
-	if (response_id == GTK_RESPONSE_CLOSE)
-		gtk_widget_destroy (GTK_WIDGET (dialog));
+	GtkWidget *msgbox;
+	gchar *utf8_file_name;
+	gchar *msg;
+	AtkObject *obj;
+	gint ret;
+
+	utf8_file_name = g_filename_to_utf8 (file_name, -1, NULL, NULL, NULL);
+	msg = g_strdup_printf (_("A file named \"%s\" already exists."),
+			       utf8_file_name);
+	g_free (utf8_file_name);
+
+	msgbox = gpdf_hig_dialog_new (
+		GTK_STOCK_DIALOG_QUESTION, msg,
+		_("Do you want to replace it with the one you are saving?"),
+		FALSE);
+
+	g_free (msg);
+
+	gtk_dialog_add_button (GTK_DIALOG (msgbox), 
+			       GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL);
+
+	gtk_dialog_add_button (GTK_DIALOG (msgbox),
+			       _("_Replace"), GTK_RESPONSE_YES);
+
+	gtk_window_set_transient_for (GTK_WINDOW (msgbox), window);
+
+	gtk_dialog_set_default_response (GTK_DIALOG (msgbox),
+					 GTK_RESPONSE_CANCEL);
+
+	obj = gtk_widget_get_accessible (msgbox);
+
+	if (GTK_IS_ACCESSIBLE (obj))
+		atk_object_set_name (obj, _("Question"));
+
+	ret = gtk_dialog_run (GTK_DIALOG (msgbox));
+	gtk_widget_destroy (msgbox);
+
+	return (ret == GTK_RESPONSE_YES);
 }
 
-/* Returns a filename in on-disk encoding (suitable for Xpdf code). */
-static gchar *
-gpdf_control_get_filename (BonoboControl *control)
+static void
+save_error_dialog (GtkWindow *window, const gchar *file_name)
 {
-	GtkWidget *filechooser;
+	char *msg;
+	GtkWidget *error_dialog;
+
+	msg = g_strdup_printf (_("The file could not be saved as \"%s\"."),
+			       file_name);
+
+	error_dialog = gpdf_hig_dialog_new (GTK_STOCK_DIALOG_ERROR,
+					    msg, NULL, FALSE);
+
+	g_free (msg);
+
+	gtk_dialog_add_button (GTK_DIALOG (error_dialog),
+			       GTK_STOCK_CLOSE,
+			       GTK_RESPONSE_CLOSE);
+
+	gtk_window_set_transient_for (GTK_WINDOW (error_dialog), window);
+
+	gtk_dialog_run (GTK_DIALOG (error_dialog));
+	gtk_widget_destroy (error_dialog);
+}
+
+static void
+gpdf_control_save_file (BonoboControl *control)
+{
+	GPdfControlPrivate *priv;
+	GtkFileFilter *pdf_filter, *all_filter;
 	gchar *pathname = NULL; 
 
-	filechooser = gtk_file_chooser_dialog_new (_("Save a Copy"),
-						   NULL,
-						   GTK_FILE_CHOOSER_ACTION_SAVE,
-						   GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
-						   GTK_STOCK_SAVE, GTK_RESPONSE_OK,
-						   NULL);
-	gtk_window_set_modal (GTK_WINDOW (filechooser), TRUE);
+	priv = GPDF_CONTROL (control)->priv;
+	priv->file_chooser = gtk_file_chooser_dialog_new (
+		_("Save a Copy"),
+		NULL, GTK_FILE_CHOOSER_ACTION_SAVE,
+		GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
+		GTK_STOCK_SAVE, GTK_RESPONSE_OK,
+		NULL);
+	gtk_window_set_modal (GTK_WINDOW (priv->file_chooser), TRUE);
+
+	pdf_filter = gtk_file_filter_new ();
+	gtk_file_filter_set_name (pdf_filter, _("PDF Documents"));
+	gtk_file_filter_add_mime_type (pdf_filter, "application/pdf");
+	gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (priv->file_chooser),
+				     pdf_filter);
+
+	all_filter = gtk_file_filter_new ();
+	gtk_file_filter_set_name (all_filter, _("All Files"));
+	gtk_file_filter_add_pattern (all_filter, "*");
+	gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (priv->file_chooser),
+				     all_filter);
+	gtk_file_chooser_set_filter (GTK_FILE_CHOOSER (priv->file_chooser),
+				     pdf_filter);
 	
-	gtk_dialog_set_default_response (GTK_DIALOG (filechooser), GTK_RESPONSE_OK);
+	gtk_dialog_set_default_response (GTK_DIALOG (priv->file_chooser),
+					 GTK_RESPONSE_OK);
 
 	bonobo_control_set_transient_for (control,
-					  GTK_WINDOW (filechooser),
+					  GTK_WINDOW (priv->file_chooser),
 					  NULL);
-	gtk_widget_show (filechooser);
+	gtk_widget_show (priv->file_chooser);
 
-	if (gtk_dialog_run(GTK_DIALOG (filechooser)) == GTK_RESPONSE_OK)
-		pathname = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (filechooser));
+	while (gtk_dialog_run (GTK_DIALOG (priv->file_chooser))
+	       == GTK_RESPONSE_OK) {
 
-	gtk_widget_destroy (filechooser); 
-	
-	return pathname; 
+		pathname = gtk_file_chooser_get_filename (
+			GTK_FILE_CHOOSER (priv->file_chooser));
+
+		if (g_file_test (pathname, G_FILE_TEST_EXISTS) &&  
+		    !overwrite_existing_file (GTK_WINDOW (priv->file_chooser),
+					      pathname))
+				continue;
+
+		if (gpdf_view_save_as (GPDF_VIEW (priv->gpdf_view), pathname))
+			break;
+		else
+			save_error_dialog (GTK_WINDOW (priv->file_chooser),
+					   pathname);    
+	}
+
+	gtk_widget_destroy (priv->file_chooser);
+	priv->file_chooser = NULL;
 }
 
 static void
 verb_SaveAs_cb (BonoboUIComponent *uic, gpointer user_data,
 		const char *cname)
 {
-	gchar *filename;
-	
 	g_return_if_fail (GPDF_IS_NON_NULL_CONTROL (user_data));
 
-	filename = gpdf_control_get_filename (BONOBO_CONTROL (user_data));
-	if (filename != NULL)
-	  gpdf_view_save_as (
-	      GPDF_VIEW (GPDF_CONTROL (user_data)->priv->gpdf_view),
-	      filename);
-	g_free (filename);
+	gpdf_control_save_file (BONOBO_CONTROL (user_data));
+}
+
+static void
+properties_response_handler (GtkDialog *dialog, gint response, gpointer data)
+{
+	gtk_widget_destroy (GTK_WIDGET (dialog));
+	GPDF_CONTROL (data)->priv->properties = NULL;
+}
+
+static gint
+properties_delete_handler (GtkWidget *widget, GdkEventAny *even,
+			   gpointer data)
+{
+	GPDF_CONTROL (data)->priv->properties = NULL;
+	return FALSE; /* destroy it in the default handler */
 }
 
 static void
 verb_FileProperties_cb (BonoboUIComponent *uic, gpointer user_data,
 			const char *cname)
 {
-	GtkWidget *dialog;
+	GPdfControlPrivate *priv;
 	GtkWidget *props;
 
 	g_return_if_fail (GPDF_IS_NON_NULL_CONTROL (user_data));
 
-	dialog = gtk_dialog_new_with_buttons (_("PDF Properties"),
-					      NULL,
-					      GTK_DIALOG_NO_SEPARATOR,
-					      GTK_STOCK_CLOSE,
-					      GTK_RESPONSE_CLOSE,
-					      NULL);
-	gtk_container_set_border_width (GTK_CONTAINER (dialog), 5);
-	gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dialog)->vbox), 2);
-	gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);
+	priv = GPDF_CONTROL (user_data)->priv;
 
-	g_signal_connect (
-		GTK_WIDGET (dialog),
-		"response",
-		G_CALLBACK (gpdf_control_properties_dialog_response_cb),
-		NULL);
+	priv->properties = gtk_dialog_new_with_buttons (_("PDF Properties"),
+							NULL,
+							GTK_DIALOG_NO_SEPARATOR,
+							GTK_STOCK_CLOSE,
+							GTK_RESPONSE_CLOSE,
+							NULL);
+	gtk_container_set_border_width (GTK_CONTAINER (priv->properties), 5);
+	gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (priv->properties)->vbox), 2);
+	gtk_window_set_resizable (GTK_WINDOW (priv->properties), FALSE);
+
+	g_signal_connect (priv->properties, "response",
+			  G_CALLBACK (properties_response_handler), user_data);
+	g_signal_connect (priv->properties, "delete-event",
+			  G_CALLBACK (properties_delete_handler), user_data);
 	bonobo_control_set_transient_for (BONOBO_CONTROL (user_data),
-					  GTK_WINDOW (dialog), NULL);
+					  GTK_WINDOW (priv->properties), NULL);
 
 	props = GTK_WIDGET (g_object_new (GPDF_TYPE_PROPERTIES_DISPLAY, NULL));
 	pdf_doc_process_properties (
@@ -1103,8 +1215,9 @@ verb_FileProperties_cb (BonoboUIComponent *uic, gpointer user_data,
 		G_OBJECT (props));
 
 	gtk_widget_show (props);
-	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), props);
-	gtk_widget_show (dialog);
+	gtk_box_pack_start_defaults (
+		GTK_BOX (GTK_DIALOG (priv->properties)->vbox), props);
+	gtk_widget_show (priv->properties);
 }
 
 static void
@@ -1405,6 +1518,7 @@ gc_set_page_cb (GPdfPageControl *page_control, gint page, GPdfControl *control)
 
 	gpdf_view_goto_page (GPDF_VIEW (control->priv->gpdf_view), page);
 	gpdf_view_scroll_to_top (GPDF_VIEW (control->priv->gpdf_view));
+	gtk_widget_grab_focus (GTK_WIDGET (control->priv->gpdf_view));
 }
 
 static void
@@ -1611,15 +1725,6 @@ gpdf_control_read_sidebar_state (GPdfControl *gpdf_control)
 			 NULL,
 			 gpdf_control);
 
-#ifndef USE_ANNOTS_VIEW
-	/*
-	 * If annots view is disabled and saved page id
-	 * is GPDF_ANNOTS_VIEW_PAGE_ID,  change it to Bookmarks page.
-	 */
-	if (!strncmp(page_id, GPDF_ANNOTS_VIEW_PAGE_ID, strlen(GPDF_ANNOTS_VIEW_PAGE_ID)))
-		page_id = GPDF_BOOKMARKS_VIEW_PAGE_ID;
-#endif
-	
 	/* Select sidebar page saved in gconf value sidebar_page */
 	gpdf_control_sidebar_page_changed_cb
 		(GPDF_SIDEBAR (gpdf_control->priv->gpdf_sidebar),
@@ -1710,7 +1815,7 @@ gpdf_control_get_bookmarks_view (GPdfControl *gpdf_control)
 	if (!priv->gpdf_bookmarks_view)
 	{
 	    priv->gpdf_bookmarks_view =
-	      gpdf_bookmarks_view_new (priv->gpdf_view);
+	      gpdf_bookmarks_view_new (gpdf_control, priv->gpdf_view);
 	    g_object_ref (G_OBJECT (priv->gpdf_bookmarks_view)); 
 	    gtk_widget_show (priv->gpdf_bookmarks_view);
 	    
@@ -1890,19 +1995,8 @@ persist_file_loading_finished_cb (GPdfPersistFile *persist_file,
 	vfs_uri = uri ? gnome_vfs_uri_new (uri) : NULL;
 
 	if (vfs_uri) {
-		/* Fix for bug #123638: unescape title */
-		const char *title =
-			gnome_vfs_uri_to_string
-			    (vfs_uri,
-			     (GnomeVFSURIHideOptions)
-			     (GNOME_VFS_URI_HIDE_TOPLEVEL_METHOD |
-			      GNOME_VFS_URI_HIDE_HOST_PORT |
-			      GNOME_VFS_URI_HIDE_HOST_NAME |
-			      GNOME_VFS_URI_HIDE_PASSWORD |
-			      GNOME_VFS_URI_HIDE_USER_NAME));
 		control->priv->title =
-			gnome_vfs_unescape_string_for_display (title);
-		g_free ((gpointer)title); 
+			gnome_vfs_uri_extract_short_name (vfs_uri);
 		gnome_vfs_uri_unref (vfs_uri);
 	} else {
 		control->priv->title = g_strdup (_("PDF Document"));
@@ -1929,16 +2023,10 @@ persist_file_loading_failed_cb (GPdfPersistFile *persist_file,
 		
 	primary = g_strdup_printf (
 		_("Loading of %s failed."),
-		gpdf_persist_file_get_current_uri (persist_file));	
+		gpdf_persist_file_get_current_uri (persist_file));
+
+	gpdf_control_private_error_dialog (control, primary, message, FALSE, TRUE);
 
-	dlg = gpdf_hig_dialog_new (GTK_STOCK_DIALOG_ERROR, primary,
-				   message, FALSE);
-	gtk_dialog_add_button (GTK_DIALOG (dlg), GTK_STOCK_OK, GTK_RESPONSE_OK);
-	g_signal_connect_swapped (G_OBJECT (dlg), "response",
-				  G_CALLBACK (gtk_widget_destroy), dlg);
-	bonobo_control_set_transient_for (BONOBO_CONTROL (control),
-					  GTK_WINDOW (dlg), NULL);
-	gtk_widget_show (dlg);
 	g_free (primary);
 
 	gpdf_control_pdf_doc_changed (control);
@@ -2237,6 +2325,8 @@ gpdf_control_sidebar_page_changed_cb (GPdfSidebar *sidebar,
 	else if (!strncmp (page_id, GPDF_ANNOTS_VIEW_PAGE_ID, strlen(GPDF_ANNOTS_VIEW_PAGE_ID)))
 		page = control->priv->gpdf_annots_view;
 #endif
+	else /* Invalid page setting: use fallback*/
+		page = control->priv->gpdf_bookmarks_view; 
 	
 	if (page)
 		gpdf_sidebar_set_content
@@ -2311,10 +2401,21 @@ gpdf_control_setup_view_widgets (GPdfControl *control)
 
 	gtk_box_pack_end (GTK_BOX (vbox), appbar, FALSE, FALSE, 0);
 
-	gpdf_view = gpdf_view_new (control);
-	control->priv->gpdf_view = gpdf_view;
+	control->priv->gpdf_view =
+	  gpdf_view =
+	    GTK_WIDGET (g_object_new (GPDF_TYPE_VIEW,
+				"aa", TRUE,	/* FIXME: a pref ? */
+				"control", control, 
+				NULL));
+
 	gtk_container_add (GTK_CONTAINER (sw), gpdf_view);
 
+	/* Workaround for 132489 */
+	gtk_scrolled_window_get_hadjustment (GTK_SCROLLED_WINDOW (sw))
+		->step_increment = 10;
+	gtk_scrolled_window_get_vadjustment (GTK_SCROLLED_WINDOW (sw))
+		->step_increment = 10;
+	
 	g_signal_connect (G_OBJECT (gpdf_view), "page_changed",
 			  G_CALLBACK (gc_page_changed_enable_page_buttons),
 			  control);
@@ -2429,6 +2530,22 @@ gpdf_control_disconnected (BonoboControl *control)
 	priv = GPDF_CONTROL (control)->priv;
 
 	gpdf_control_save_sidebar_state (GPDF_CONTROL (control));
+
+	if (priv->file_chooser) {
+		gtk_widget_destroy (priv->file_chooser);
+		priv->file_chooser = NULL;
+	}
+
+	if (priv->properties) {
+		gtk_widget_destroy (priv->properties);
+		priv->properties = NULL;
+	}
+
+	if (priv->print_dialog) {
+		gtk_widget_destroy (priv->print_dialog);
+		priv->print_dialog = NULL;
+	}
+
 	BONOBO_CALL_PARENT (BONOBO_CONTROL_CLASS, disconnected, (control));
 }
 
