@@ -25,7 +25,7 @@
 #include "gpdf-util.h"
 #include "gpdf-link-canvas-item.h"
 #include "gpdf-g-switch.h"
-#  include <libgnome/gnome-i18n.h>
+#  include <glib/gi18n.h>
 #  include <libgnome/gnome-macros.h>
 #  include "gpdf-marshal.h"
 #include "gpdf-g-switch.h"
@@ -148,7 +148,8 @@ gpdf_link_canvas_item_set_link (GPdfLinkCanvasItem *link_item, Link *link)
 
 	link_item->priv->link = link;
 
-	link->getBorder (&x1, &y1, &x2, &y2, &w);
+	w = 1.0;
+	link->getRect (&x1, &y1, &x2, &y2);
 	g_object_set (G_OBJECT (link_item),
 		      "x1", x1, "y1", y1,
 		      "x2", x2, "y2", y2,
