@@ -47,6 +47,7 @@
 #include "GfxState.h"
 #include "Thumb.h"
 #include "gpdf-view.h"
+#include "gpdf-control-private.h"
 
 BEGIN_EXTERN_C
 
@@ -56,7 +57,7 @@ BEGIN_EXTERN_C
 /* FIXME: Put this values in preferences ? */
 #define GPDF_THUMBNAILS_PER_ROW		2
 #define GPDF_THUMBNAILS_ENLIGTH_FACTOR	35.0
-#define GPDF_EMPTY_THUMB_IMAGE		GNOMEICONDIR "/gpdf/empty-thumb.png"
+#define GPDF_DEFAULT_THUMB_WIDTH	76
 
 typedef struct _GPdfThumbnail {
 	GnomeCanvasItem *group;
@@ -91,7 +92,10 @@ struct _GPdfThumbnailsViewPrivate {
 	GtkWidget *popup_menu;
 	
 	gint current_page;
-	double zoom; 
+	double zoom;
+	guint current_gen_thumb;
+	guint idle_id; 
+	gulong idle_dcon_id; 
 
 	gint clicked_x, clicked_y;
 
@@ -110,7 +114,6 @@ typedef enum _GPdfThumbnailsDragAction {
 } GPdfThumbnailsDragAction; 
 
 static void	gpdf_thumbnails_view_class_init	      		     (GPdfThumbnailsViewClass*); 
-static void	gpdf_thumbnails_view_class_finalize   		     (gpointer, gpointer); 
 static void	gpdf_thumbnails_view_dispose	      		     (GObject*); 
 static void	gpdf_thumbnails_view_finalize	      		     (GObject*); 
 static void	gpdf_thumbnails_view_instance_init    		     (GPdfThumbnailsView*);
@@ -119,19 +122,28 @@ static void	gpdf_thumbnails_view_get_page_canvas_regions         (GPdfThumbnails
 							              double *px2, double *py2,
 							              int *cx, int *cy,
 							              double *vw, double *vh); 
+#if CAN_GENERATE_THUMBNAILS
 static void	gpdf_thumbnails_view_popup_menu_item_generate_cb     (GtkMenuItem*,
 								      GPdfThumbnailsView*); 
 static void	gpdf_thumbnails_view_popup_menu_item_generate_all_cb (GtkMenuItem*,
-								      GPdfThumbnailsView*); 
+								      GPdfThumbnailsView*);
+#endif
 
 #define GPDF_IS_NON_NULL_THUMBNAILS_VIEW(obj) \
 		(((obj) != NULL) && (GPDF_IS_THUMBNAILS_VIEW ((obj))))
 
 enum {
 	THUMBNAIL_SELECTED_SIGNAL = 0,
+	READY_SIGNAL, 
 	LAST_SIGNAL
 };
 
+enum {
+	PROP_0,
+	PROP_VIEW,
+	PROP_CONTROL
+};
+
 #define POPUP_MENU_GENERATE_ITEM		"Generate Thumbnail"
 #define POPUP_MENU_GENERATE_ITEM_TIP		"Generate Thumbnail image for current page"
 #define POPUP_MENU_GENERATE_ALL_ITEM		"Generate All Thumbnails"
@@ -157,6 +169,31 @@ static guint gpdf_thumbnails_view_signals [LAST_SIGNAL];
 
 GPDF_CLASS_BOILERPLATE(GPdfThumbnailsView, gpdf_thumbnails_view, GtkScrolledWindow, GTK_TYPE_SCROLLED_WINDOW); 
 
+static void
+gpdf_thumbnails_view_set_property (GObject *object, guint param_id,
+				  const GValue *value, GParamSpec *pspec)
+{
+	GPdfThumbnailsView *thumbnails_view;
+
+	g_return_if_fail (GPDF_IS_THUMBNAILS_VIEW (object));
+
+	thumbnails_view = GPDF_THUMBNAILS_VIEW (object);
+
+	switch (param_id) {
+	case PROP_VIEW:
+		thumbnails_view->priv->gpdf_view = 
+		  GPDF_VIEW (g_value_get_object (value));
+		break;
+	case PROP_CONTROL:
+		thumbnails_view->priv->parent = 
+		  GPDF_CONTROL (g_value_get_object (value));
+		break;
+	default:
+		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
+		break;
+	}
+}
+
 static void
 gpdf_thumbnails_view_class_init (GPdfThumbnailsViewClass *klass)
 {
@@ -168,6 +205,25 @@ gpdf_thumbnails_view_class_init (GPdfThumbnailsViewClass *klass)
 
 	gobject_class->dispose = gpdf_thumbnails_view_dispose;
 	gobject_class->finalize = gpdf_thumbnails_view_finalize;
+	gobject_class->set_property = gpdf_thumbnails_view_set_property;
+
+	g_object_class_install_property (
+		gobject_class, PROP_VIEW,
+		g_param_spec_object (
+			"view",
+			"Parent view",
+			"Parent view",
+			GPDF_TYPE_VIEW, 
+		        (GParamFlags)(G_PARAM_CONSTRUCT_ONLY | G_PARAM_WRITABLE)));
+
+	g_object_class_install_property (
+		gobject_class, PROP_CONTROL,
+		g_param_spec_object (
+			"control",
+			"Parent control",
+			"Parent control",
+			GPDF_TYPE_CONTROL, 
+		        (GParamFlags)(G_PARAM_CONSTRUCT_ONLY | G_PARAM_WRITABLE)));
 
 	gpdf_thumbnails_view_signals [THUMBNAIL_SELECTED_SIGNAL] =
 	  g_signal_new (
@@ -181,6 +237,16 @@ gpdf_thumbnails_view_class_init (GPdfThumbnailsViewClass *klass)
 	    G_TYPE_INT, G_TYPE_INT, /* x, y */
 	    G_TYPE_INT, G_TYPE_INT, /* w, h */
 	    G_TYPE_INT); /* page */
+
+	gpdf_thumbnails_view_signals [READY_SIGNAL] =
+	  g_signal_new (
+	    "ready",
+	    G_TYPE_FROM_CLASS (gobject_class),
+	    G_SIGNAL_RUN_LAST,
+	    G_STRUCT_OFFSET (GPdfThumbnailsViewClass, ready),
+	    NULL, NULL,
+	    gpdf_marshal_VOID__VOID,
+	    G_TYPE_NONE, 0);
 }
 
 static void
@@ -195,11 +261,11 @@ gpdf_thumbnails_view_emit_thumbnail_selected (GPdfThumbnailsView *thumbnails_vie
 }
 
 static void
-gpdf_thumbnails_view_init (GPdfThumbnailsView *thumbnails_view)
+gpdf_thumbnails_view_emit_ready (GPdfThumbnailsView *thumbnails_view)
 {
-	GtkScrolledWindow *window;
-
-	window = GTK_SCROLLED_WINDOW (thumbnails_view);
+	g_signal_emit (G_OBJECT (thumbnails_view),
+		       gpdf_thumbnails_view_signals [READY_SIGNAL],
+		       0);
 }
 
 static void
@@ -259,6 +325,12 @@ gpdf_thumbnails_view_page_box_coord(GPdfThumbnailsView *thumbnails_view,
 	real_width = gdk_pixbuf_get_width (priv->thumbs[page -1].pixbuf); 
 	real_height = gdk_pixbuf_get_height (priv->thumbs[page -1].pixbuf); 
 
+	/* Update page size if guess was wrong */
+	if (priv->max_page_width < real_width)
+		priv->max_page_width = real_width; 
+	if (priv->max_page_height < real_height)
+		priv->max_page_height = real_height; 
+
 	if (x) {
 		*x = GPDF_THUMBNAILS_MARGIN_LEFT +
 			((page -1) % priv->page_per_row) *
@@ -267,7 +339,6 @@ gpdf_thumbnails_view_page_box_coord(GPdfThumbnailsView *thumbnails_view,
 			 GPDF_THUMBNAILS_INNER_MARGIN_LEFT +
 			 GPDF_THUMBNAILS_SHADOW_WIDTH);
 
-		g_assert (priv->max_page_width >= real_width); 
 		*x += (priv->max_page_width -real_width)/2;
 	}
 
@@ -310,7 +381,7 @@ gpdf_thumbnails_view_scroll_region_coord(GPdfThumbnailsView *thumbnails_view,
 				      GPDF_THUMBNAILS_INNER_MARGIN_RIGHT) +
 		(priv->page_per_row -1) * GPDF_THUMBNAILS_COL_SEP;
 
-	npage_h = (abs(page / priv->page_per_row) + (page % priv->page_per_row ? 1 : 0));
+	npage_h = (labs(page / priv->page_per_row) + (page % priv->page_per_row ? 1 : 0));
 	*y = GPDF_THUMBNAILS_MARGIN_TOP + GPDF_THUMBNAILS_MARGIN_BOTTOM +
 		npage_h * (priv->max_page_height +
 			   GPDF_THUMBNAILS_INNER_MARGIN_TOP +
@@ -327,13 +398,12 @@ gpdf_thumbnails_view_set_scroll_region(GPdfThumbnailsView *thumbnails_view)
 	GPdfThumbnailsViewPrivate *priv;
 	guint page; 
 	double x, y;
-	double x2, y2;
     
 	g_return_if_fail (GPDF_IS_NON_NULL_THUMBNAILS_VIEW (thumbnails_view));
 
 	priv = thumbnails_view->priv;
 
-	for (page = 1; page <= priv->num_pages; page++) {
+	for (page = 1; page <= priv->current_gen_thumb; page++) {
 		gpdf_thumbnails_view_page_box_coord (thumbnails_view,
 						     page,
 						     &x, &y,
@@ -361,7 +431,7 @@ gpdf_thumbnails_view_get_page_outline_coords (GPdfThumbnailsView *thumbnails_vie
 	GPdfThumbnailsViewPrivate *priv;
 	GPdfThumbnail *thumb;
 	double px1, px2, py1, py2;
-	double vw, vh;
+	double vw = 0.0, vh = 0.0;
 	int cx, cy;
 	double page_width, page_height;
 	int width, height; 
@@ -502,7 +572,6 @@ gpdf_thumbnails_view_page_changed_cb (GPdfView *gpdf_view,
 {
 	GPdfThumbnailsView *thumbnails_view;
 	GPdfThumbnailsViewPrivate *priv;
-	GtkTreeModel *model;
 	double item_x, item_y;
 
 	thumbnails_view = GPDF_THUMBNAILS_VIEW (user_data);
@@ -534,7 +603,6 @@ gpdf_thumbnails_view_zoom_changed_cb (GPdfView *gpdf_view,
 {
 	GPdfThumbnailsView *thumbnails_view;
 	GPdfThumbnailsViewPrivate *priv;
-	GtkTreeModel *model;
 
 	thumbnails_view = GPDF_THUMBNAILS_VIEW (user_data);
 
@@ -555,15 +623,16 @@ GtkWidget *
 gpdf_thumbnails_view_new (GPdfControl *parent, GtkWidget *gpdf_view)
 {
 	GPdfThumbnailsView *thumbnails_view;
-	GPdfThumbnailsViewPrivate *priv;
 
 	thumbnails_view =
 	  GPDF_THUMBNAILS_VIEW (g_object_new (
-		  GPDF_TYPE_THUMBNAILS_VIEW, NULL));
+		  GPDF_TYPE_THUMBNAILS_VIEW, 
+		  "view", (gpointer)gpdf_view,
+		  "control", (gpointer)parent, 
+		  NULL));
 
-	thumbnails_view->priv->parent = parent;
-	thumbnails_view->priv->gpdf_view = GPDF_VIEW (gpdf_view);
 	g_object_ref (G_OBJECT (thumbnails_view->priv->gpdf_view)); 
+	g_object_ref (G_OBJECT (thumbnails_view->priv->parent)); 
 	gpdf_thumbnails_view_construct (thumbnails_view); 
 	
 	g_signal_connect (G_OBJECT (gpdf_view), "page_changed",
@@ -604,6 +673,10 @@ gpdf_thumbnails_view_dispose (GObject *object)
 		priv->popup_menu = NULL;
 		g_free (priv->popup_menu_uiinfo);
 	}
+	if (priv->parent) {
+		g_object_unref (priv->parent);
+		priv->parent = NULL; 
+	}
 
 	G_OBJECT_CLASS (parent_class)->dispose (object);
 }
@@ -648,9 +721,7 @@ gpdf_thumbnails_view_get_page_canvas_regions (GPdfThumbnailsView *thumbnails_vie
 					      int *cx, int *cy,
 					      double *vw, double *vh)
 {
-	GPdfThumbnailsViewPrivate *priv;
 	GPdfView *gpdf_view;
-	double page_size; 
 
 	g_return_if_fail (GPDF_IS_NON_NULL_THUMBNAILS_VIEW (thumbnails_view));
 
@@ -796,7 +867,7 @@ gpdf_thumbnails_view_canvas_outline_event_cb (GnomeCanvasItem *outline,
 		/* When leaving area restore cursor */
 		display = gtk_widget_get_display (priv->canvas_widget);
 		cursor = gdk_cursor_new_for_display (display, GDK_LEFT_PTR);
-		gdk_window = gtk_widget_get_parent_window (priv->canvas_widget);
+		gdk_window = GTK_WIDGET (priv->canvas_widget)->window;
 		gdk_window_set_cursor (gdk_window, cursor);
 		gdk_cursor_unref (cursor);
 		gdk_flush();
@@ -881,7 +952,7 @@ gpdf_thumbnails_view_canvas_outline_event_cb (GnomeCanvasItem *outline,
 		else {
 			cursor = gdk_cursor_new_for_display (display, GDK_LEFT_PTR);
 		}
-		gdk_window = gtk_widget_get_parent_window (priv->canvas_widget);
+		gdk_window = GTK_WIDGET (priv->canvas_widget)->window;
 		gdk_window_set_cursor (gdk_window, cursor); 
 		gdk_cursor_unref (cursor);
 		gdk_flush();
@@ -893,17 +964,10 @@ gpdf_thumbnails_view_canvas_outline_event_cb (GnomeCanvasItem *outline,
 			hadj = GTK_LAYOUT (priv->gpdf_view)->hadjustment; 
 			vadj = GTK_LAYOUT (priv->gpdf_view)->vadjustment; 
 			
-/* 			g_message ("h adj: %.1f < %.1f < %.1f  size/step/pstep = %.1f/%.1f/%.1f", */
-/* 				   hadj->lower, hadj->value, hadj->upper, hadj->page_size, */
-/* 				   hadj->step_increment, hadj->page_increment); */
-/* 			g_message ("v adj: %.1f < %.1f < %.1f  size/step/pstep = %.1f/%.1f/%.1f", */
-/* 				   vadj->lower, vadj->value, vadj->upper, vadj->page_size, */
-/* 				   vadj->step_increment, vadj->page_increment);  */
-			
 			hval = gtk_adjustment_get_value(hadj) +
-				((outline_x -x) * priv->zoom * (hadj->upper -hadj->lower) / view_width); 
+				((outline_x -x) * (hadj->upper -hadj->lower) / view_width); 
 			vval = gtk_adjustment_get_value(vadj) +
-				((outline_y -y) * priv->zoom * (vadj->upper -vadj->lower) / view_height);
+				((outline_y -y) * (vadj->upper -vadj->lower) / view_height);
 
 			gtk_adjustment_set_value(hadj, MIN (MAX (hval, hadj->lower), hadj->upper - hadj->page_size));
 			gtk_adjustment_set_value(vadj, MIN (MAX (vval, vadj->lower), vadj->upper - vadj->page_size)); 
@@ -978,10 +1042,10 @@ gpdf_thumbnails_view_create_thumbnail (GPdfThumbnailsView *thumbnails_view,
 	GPdfThumbnailsViewPrivate *priv; 
 	GnomeCanvasItem *group;
 	GnomeCanvasItem *item;
+	double dx, dy, dw, dh;
 	int width, height; 
 	double outerwidth, outerheight;
 	double shadow_x, shadow_y;
-	double text_x, text_y;
 	gchar *text;
 
 	g_return_val_if_fail (GPDF_IS_NON_NULL_THUMBNAILS_VIEW (thumbnails_view),
@@ -990,17 +1054,28 @@ gpdf_thumbnails_view_create_thumbnail (GPdfThumbnailsView *thumbnails_view,
 	priv = thumbnails_view->priv;
 
 	thumbnail->pixbuf = pixbuf;
-	
-	group = gnome_canvas_item_new(priv->root,
-				      gnome_canvas_group_get_type(),
-				      "x", (double)0,
-				      "y", (double)0, 
-				      NULL);
-	thumbnail->group = group;
 
 	width = gdk_pixbuf_get_width (pixbuf); 
 	height = gdk_pixbuf_get_height (pixbuf); 
 
+	/*
+	 * Just a guess for now, in order to have
+	 * an approximate layout while creating
+	 * thumbs
+	 */
+	dw = (double)width;
+	dh = (double)height;
+	gpdf_thumbnails_view_page_box_coord(thumbnails_view,
+					    page, 
+					    &dx, &dy,
+					    &dw, &dh);
+	group = gnome_canvas_item_new(priv->root,
+				      gnome_canvas_group_get_type(),
+				      "x", dx,
+				      "y", dy, 
+				      NULL);
+	thumbnail->group = group;
+
 	outerwidth = width +
 	  	GPDF_THUMBNAILS_INNER_MARGIN_LEFT +
 		GPDF_THUMBNAILS_INNER_MARGIN_RIGHT +
@@ -1105,17 +1180,13 @@ gpdf_thumbnails_view_create_thumbnail_image (GPdfThumbnailsView *thumbnails_view
 					     gint *width, gint *height)
 {
 	GPdfThumbnailsViewPrivate *priv; 
-	gint lwidth = 76, lheight = 99;
-	GfxColorSpace *colorspace;
+	gint lwidth = GPDF_DEFAULT_THUMB_WIDTH;
 	GdkColorspace gdk_colorspace = GDK_COLORSPACE_RGB;
 	gboolean has_alpha = FALSE; 
 	gint bitsPerComponent = 8;
 	GdkPixbuf *pixbuf;
 	gint dest_width, dest_height;
 	gdouble page_ratio;
-	GtkStyle *gtkstyle;
-	GdkGC *gdkgc;
-	GdkGCValues gcvals;
 	
 	g_return_if_fail (GPDF_IS_NON_NULL_THUMBNAILS_VIEW (thumbnails_view)); 
 	
@@ -1175,9 +1246,9 @@ gpdf_thumbnails_view_render_thumbnail_image (GPdfThumbnailsView *thumbnails_view
 	pixbuf = gdk_pixbuf_new_from_data(data,
 					  GDK_COLORSPACE_RGB,
 					  FALSE,
-					  bitsPerComponent,
+					  8,
 					  width, height,
-					  width*3,
+					  width * 3,
 					  NULL, NULL);
 
 	/* Create the thumbnail itself */
@@ -1195,104 +1266,142 @@ static gboolean
 gpdf_thumbnails_view_populate_idle (GPdfThumbnailsView *thumbnails_view)
 {
 	GPdfThumbnailsViewPrivate *priv;
-	gint num_pages;
-	guint page;
 	Page *the_page;
 	Object the_thumb;
-	Thumb *thumb;
-	XRef *xref;
+	Thumb *thumb = NULL;
+	gboolean have_ethumbs = FALSE;
+	gchar *status = NULL;
+	double ratio = 0.0; 
 	
 	g_return_val_if_fail (GPDF_IS_NON_NULL_THUMBNAILS_VIEW (thumbnails_view),
 			      TRUE);
 	
 	priv = thumbnails_view->priv;
+	ratio = ((double)priv->current_gen_thumb/(double)priv->num_pages); 
+	status = g_strdup_printf ("Processing thumbnail #%d/%d: %3d %%",
+				  priv->current_gen_thumb,
+				  priv->num_pages, 
+				  (int)(ratio*100.0));
+	gpdf_control_private_set_status (priv->parent, status);
+	gpdf_control_private_set_fraction (priv->parent, ratio); 
+	g_free (status); 
 
-	/* Get number of pages */
-	priv->num_pages = priv->pdf_doc->getNumPages ();
-
-	/* Allocate thumbs table */
-	priv->thumbs =(GPdfThumbnail *)g_new0 (GPdfThumbnail, priv->num_pages);
-
-	/* Init page max sizes */
-	priv->max_page_width = 0;
-	priv->max_page_height = 0; 	
-
-	/* Create/Read thumbnails */
-	for (page = 1; page <= priv->num_pages; page++) {
-
-		/* Get the PDF objects for fetching thumb */
-		the_page = priv->pdf_doc->getCatalog ()->getPage (page);
-		xref = priv->pdf_doc->getXRef ();
-		the_page->getThumb(&the_thumb);
-
+	/* Get the PDF objects for fetching thumb */
+	the_page = priv->pdf_doc->getCatalog ()->getPage (++(priv->current_gen_thumb));
+	the_page->getThumb(&the_thumb);
+	
+	/*
+	 * If no thumb object available in PDF, render each page in
+	 * thumbnail mode, else fetch thumbnail object and render image
+	 * data.
+	 */
+	if (!(the_thumb.isNull () || the_thumb.isNone())) {
+		/* Build the thumbnail object */
+		thumb = new Thumb(priv->pdf_doc->getXRef (),
+				  &the_thumb);
+		
+		have_ethumbs = thumb->ok();
+	}
+	
+	if (have_ethumbs) {
+		/* Keep maximum width/height to compute layout later */
+		priv->max_page_width = MAX ((guint)thumb->getWidth(), priv->max_page_width);
+		priv->max_page_height = MAX ((guint)thumb->getHeight(), priv->max_page_height);
+		
+		gpdf_thumbnails_view_render_thumbnail_image (thumbnails_view,
+							     priv->current_gen_thumb, 
+							     thumb->getWidth(),
+							     thumb->getHeight(),
+							     thumb->getColorSpace(),
+							     thumb->getBitsPerComponent(),
+							     thumb);
+	}
+	else {
+		GtkWidget *item;
+		guint width, height;
+		
 		/*
-		 * If no thumb object available in PDF, render each page in
-		 * thumbnail mode, else fetch thumbnail object and render image
-		 * data.
+		 * FIXME: Could perhaps render only displayed thumbs. This process may
+		 * be time consuming.
 		 */
-		if (the_thumb.isNull () || the_thumb.isNone()) {
-			GtkWidget *item; 
-			guint width, height;
-			
-			/*
-			 * FIXME: Could perhaps render only displayed thumbs. This process may
-			 * be time consuming.
-			 */
-			gpdf_thumbnails_view_create_thumbnail_image (thumbnails_view, the_page, page,
-								     (gint*)&width, (gint*)&height);
-
-			/* Keep maximum width/height to compute layout later */
-			priv->max_page_width = MAX (width, priv->max_page_width);
-			priv->max_page_height = MAX (height, priv->max_page_height);
-
+		gpdf_thumbnails_view_create_thumbnail_image (thumbnails_view, the_page,
+							     priv->current_gen_thumb, 
+							     (gint*)&width, (gint*)&height);
+		
+		/* Keep maximum width/height to compute layout later */
+		priv->max_page_width = MAX (width, priv->max_page_width);
+		priv->max_page_height = MAX (height, priv->max_page_height);
+		
 #if CAN_GENERATE_THUMBNAILS
-			item = priv->popup_menu_uiinfo[POPUP_MENU_GENERATE_INDEX].widget;
-			gtk_widget_set_sensitive (item, TRUE); 
-			item = priv->popup_menu_uiinfo[POPUP_MENU_GENERATE_ALL_INDEX].widget;
-			gtk_widget_set_sensitive (item, TRUE); 
+		item = priv->popup_menu_uiinfo[POPUP_MENU_GENERATE_INDEX].widget;
+		if (GTK_IS_WIDGET (item))
+		  gtk_widget_set_sensitive (item, TRUE); 
+		item = priv->popup_menu_uiinfo[POPUP_MENU_GENERATE_ALL_INDEX].widget;
+		if (GTK_IS_WIDGET (item))
+		  gtk_widget_set_sensitive (item, TRUE); 
 #endif
-		}
-		else
-		{
-			/* Build the thumbnail object */
-			thumb = new Thumb(priv->pdf_doc->getXRef (),
-					  &the_thumb);
-
-			/* Keep maximum width/height to compute layout later */
-			priv->max_page_width = MAX ((guint)thumb->getWidth(), priv->max_page_width);
-			priv->max_page_height = MAX ((guint)thumb->getHeight(), priv->max_page_height);
-			
-			gpdf_thumbnails_view_render_thumbnail_image (thumbnails_view, page,
-								     thumb->getWidth(), thumb->getHeight(),
-								     thumb->getColorSpace(),
-								     thumb->getBitsPerComponent(),
-								     thumb);
-		}
 	}
 
-	/* Now width and height are known, set scroll region and relayout */
+	/* Width and height are best known, set scroll region and relayout */
 	gpdf_thumbnails_view_set_scroll_region(thumbnails_view);
-	
-	/* All thumbs are setup, let listen page canvas adjustments to update page outline */
-	g_signal_connect (GTK_LAYOUT (priv->gpdf_view)->hadjustment, "value_changed",
-			  G_CALLBACK (gpdf_thumbnails_view_page_layout_adjustment_changed),
-			  thumbnails_view); 
-	g_signal_connect (GTK_LAYOUT (priv->gpdf_view)->hadjustment, "changed",
-			  G_CALLBACK (gpdf_thumbnails_view_page_layout_adjustment_changed),
-			  thumbnails_view); 
-	g_signal_connect (GTK_LAYOUT (priv->gpdf_view)->vadjustment, "value_changed",
-			  G_CALLBACK (gpdf_thumbnails_view_page_layout_adjustment_changed),
-			  thumbnails_view); 
-	g_signal_connect (GTK_LAYOUT (priv->gpdf_view)->vadjustment, "changed",
-			  G_CALLBACK (gpdf_thumbnails_view_page_layout_adjustment_changed),
-			  thumbnails_view); 
-	
-	/* Select first page */
-	gpdf_thumbnails_view_set_selected_page (thumbnails_view, 1);
-	
-	priv->thumbs_initialized = TRUE;
+		
+	if (priv->current_gen_thumb == priv->num_pages)
+	{
+		GdkDisplay *display;
+		GdkWindow *parent_window;
+		
+		/* All thumbs are setup, let listen page canvas adjustments to update page outline */
+		g_signal_connect (GTK_LAYOUT (priv->gpdf_view)->hadjustment, "value_changed",
+				  G_CALLBACK (gpdf_thumbnails_view_page_layout_adjustment_changed),
+				  thumbnails_view); 
+		g_signal_connect (GTK_LAYOUT (priv->gpdf_view)->hadjustment, "changed",
+				  G_CALLBACK (gpdf_thumbnails_view_page_layout_adjustment_changed),
+				  thumbnails_view); 
+		g_signal_connect (GTK_LAYOUT (priv->gpdf_view)->vadjustment, "value_changed",
+				  G_CALLBACK (gpdf_thumbnails_view_page_layout_adjustment_changed),
+				  thumbnails_view); 
+		g_signal_connect (GTK_LAYOUT (priv->gpdf_view)->vadjustment, "changed",
+				  G_CALLBACK (gpdf_thumbnails_view_page_layout_adjustment_changed),
+				  thumbnails_view); 
+		
+		/* Select first page */
+		gpdf_thumbnails_view_set_selected_page (thumbnails_view, 1);
+		
+		priv->thumbs_initialized = TRUE;
+		
+		display = gtk_widget_get_display (priv->canvas_widget);
+		parent_window = GTK_WIDGET (priv->canvas_widget)->window;
+		if (GDK_IS_WINDOW (parent_window))
+			gdk_window_set_cursor (parent_window, NULL); 
+		gdk_flush();
 
-	return FALSE; 
+		gpdf_control_private_clear_stack (priv->parent);
+		gpdf_control_private_set_fraction (priv->parent, 0.0); 
+		g_signal_handler_disconnect (G_OBJECT (priv->parent), priv->idle_dcon_id); 
+		gpdf_thumbnails_view_emit_ready (thumbnails_view);
+	}
+	else {
+		GdkDisplay *display;
+		GdkCursor *cursor;
+		GdkWindow *parent_window;
+		
+		/* Set watch cursor while view not ready */
+		display = gtk_widget_get_display (priv->canvas_widget);
+		cursor = gdk_cursor_new_for_display (display, GDK_WATCH);
+		parent_window = GTK_WIDGET (priv->canvas_widget)->window;
+		if (GDK_IS_WINDOW (parent_window))
+			gdk_window_set_cursor (parent_window, cursor); 
+		gdk_cursor_unref (cursor);
+		gdk_flush();
+	}
+
+	return (priv->current_gen_thumb != priv->num_pages); 
+}
+
+static void
+disconnected_handler (gpointer control, gpointer data)
+{
+	g_source_remove (GPOINTER_TO_UINT (data));
 }
 
 void
@@ -1309,15 +1418,40 @@ gpdf_thumbnails_view_set_pdf_doc (GPdfThumbnailsView *thumbnails_view,
     {
 	    priv->pdf_doc = pdf_doc;
 
+	    /* Get number of pages */
+	    priv->num_pages = priv->pdf_doc->getNumPages ();
+
+	    /* Allocate thumbs table */
+	    priv->thumbs =(GPdfThumbnail *)g_new0 (GPdfThumbnail, priv->num_pages);
+
+	    /*
+	     * Init page max sizes: We try a guess in order to have thumbs created
+	     * in an approximate place. When READY signal will be emitted the whole
+	     * layout will be recomputed with exact values.
+	     */
+	    priv->max_page_width = 45;
+	    priv->max_page_height = 90;
+
+	    /* Init index of last generated thumb */
+	    priv->current_gen_thumb = 0;
+
 	    /*
 	     * As thumb computation may be a long process
 	     * we make it as backgroud task.
 	     */
-	    g_idle_add ((GSourceFunc)gpdf_thumbnails_view_populate_idle,
-			thumbnails_view);
+	    priv->idle_id = g_idle_add_full (G_PRIORITY_DEFAULT_IDLE +1,
+					     (GSourceFunc)gpdf_thumbnails_view_populate_idle,
+					     thumbnails_view, 
+					     NULL);
+
+	    priv->idle_dcon_id =
+		    g_signal_connect (G_OBJECT (priv->parent), "disconnected",
+				      G_CALLBACK (disconnected_handler),
+				      GUINT_TO_POINTER (priv->idle_id));
     }
 }
 
+#if CAN_GENERATE_THUMBNAILS
 static void
 gpdf_thumbnails_view_popup_menu_item_generate_cb (GtkMenuItem *item,
 						  GPdfThumbnailsView *thumbnails_view)
@@ -1329,8 +1463,8 @@ gpdf_thumbnails_view_popup_menu_item_generate_cb (GtkMenuItem *item,
 	priv = thumbnails_view->priv;
 	
 	gpdf_control_private_warn_dialog (priv->parent, 
-					  _("Not yet implemented !"),
-					  _("Thumbnail Generation feature not yet implemented ... Sorry."),
+					  _("Not yet implemented!"),
+					  _("Thumbnail Generation feature not yet implemented... Sorry."),
 					  TRUE);					  
 }
 
@@ -1345,20 +1479,21 @@ gpdf_thumbnails_view_popup_menu_item_generate_all_cb (GtkMenuItem *item,
 	priv = thumbnails_view->priv;
 	
 	gpdf_control_private_warn_dialog (priv->parent, 
-					  _("Not yet implemented !"),
-					  _("All Thumbnails Generation feature not yet implemented ... Sorry."),
+					  _("Not yet implemented!"),
+					  _("All Thumbnails Generation feature not yet implemented... Sorry."),
 					  TRUE);					  
 }
+#endif
 
 GtkWidget*
 gpdf_thumbnails_view_get_tools_menu (GPdfThumbnailsView *thumbnails_view)
 {
+#if CAN_GENERATE_THUMBNAILS
 	GPdfThumbnailsViewPrivate *priv;
 	GtkWidget *item; 
 	
 	g_return_val_if_fail (GPDF_IS_NON_NULL_THUMBNAILS_VIEW (thumbnails_view), NULL);
 	
-#if CAN_GENERATE_THUMBNAILS
 	priv = thumbnails_view->priv;
 	
 	if (!priv->popup_menu) {
